@@ -31,7 +31,7 @@
 
 
 /*
- * $Id: machine.h,v 1.41 2006/03/27 23:09:21 abe Exp $
+ * $Id: machine.h,v 1.43 2010/07/29 16:02:44 abe Exp $
  */
 
 
@@ -62,7 +62,7 @@
  * For gcc's sake, some redefinitions after including <sys/types.h> insure
  * the off64_t and offset_t types are aligned on an 8 byte boundary.
  */
- 
+
 #  if	AIXV>=4300
 #define	_KERNEL	1
 #  endif	/* AIXV>=4300 */
@@ -500,6 +500,13 @@ typedef	long long	aligned_offset_t  __attribute__ ((aligned (8)));
 /* #define	HASSNODE	1 */
 
 
+/*
+ * HASTASKS is defined for those dialects that have task reporting support.
+ */
+
+/* #define	HASTASKS	1 */
+
+
 /*
  * HASSOOPT, HASSOSTATE and HASTCPOPT define the availability of information
  * on socket options (SO_* symbols), socket states (SS_* symbols) and TCP
@@ -552,6 +559,15 @@ typedef	long long	aligned_offset_t  __attribute__ ((aligned (8)));
 /* #define	HASTCPTPIW	1 */
 
 
+/*
+ * HASTCPUDPSTATE is defined for dialects that have TCP and UDP state
+ * support -- i.e., for the "-stcp|udp:state" option and its associated
+ * speed improvements.
+ */
+
+#define	HASTCPUDPSTATE	1
+
+
 /*
  * HASTMPNODE is defined for those dialects that have tmpnodes.
  */
