@@ -52,23 +52,6 @@
 #include <sys/_lock.h>
 #undef _KERNEL
 #endif         /* FREEBSDV>=13000 */
-#   if	FREEBSDV<6020
-#    if	defined(__alpha__)
-/*
- * For Alpha below 6.2, #include <machine/pcpu.h> before #define'ing _KERNEL.
- * Then #define PCPU_MD_FIELDS independently.  This hack avoids a compiler
- * complaint about register use.
- */
-
-#include <machine/pcpu.h>
-#define PCPU_MD_FIELDS                                                  \
-	struct alpha_pcb pc_idlepcb;            /* pcb for idling */    \
-	u_int64_t       pc_idlepcbphys;         /* pa of pc_idlepcb */  \
-	u_int64_t       pc_pending_ipis;        /* pending IPI's */     \
-	u_int32_t       pc_next_asn;            /* next ASN to alloc */ \
-	u_int32_t       pc_current_asngen       /* ASN rollover check */
-#    endif	/* defined(__alpha__) */
-#   endif	/* FREEBSDV<6020 */
 #define	_KERNEL	1
 
 #  if	defined(HAS_VM_MEMATTR_T)
@@ -133,6 +116,8 @@ typedef	struct device	*device_t;
 #define	NFS
 #define m_stat	mnt_stat
 
+struct statfs;
+extern int	statfs(const char *, struct statfs *);
 #define	_KERNEL
 
 #include <sys/mount.h>
@@ -183,7 +168,6 @@ int     getmntinfo(struct statfs **, int);
 #undef	_KVM_VNODE
 # endif	/* defined(HAS_KVM_VNODE) */
 
-#include <net/raw_cb.h>
 #include <sys/domain.h>
 #define	pmap	RPC_pmap
 #include <rpc/rpc.h>
@@ -237,7 +221,9 @@ struct vop_setextattr_args;
 #include <nfsclient/nfs.h>
 #include <nfsclient/nfsnode.h>
 
+#define syscall_args __bad_syscall_args
 #include <sys/proc.h>
+#undef __bad_syscall_args
 #include <kvm.h>
 #undef	TRUE
 #undef	FALSE
@@ -261,20 +247,10 @@ struct vop_generic_args;
 #undef	KERNEL
 # endif	/* defined(HASNULLFS) */
 
-# if	defined(HASPROCFS)
-#include <machine/reg.h>
-
-#define	PNSIZ		5
-# endif	/* defined(HASPROCFS) */
-
 # if	defined(HASPSEUDOFS)
 #include <fs/pseudofs/pseudofs.h>
 # endif	/* defined(HASPSEUDOFS) */
 
-# if	defined(HAS_ZFS)
-#include "dzfs.h"
-# endif	/* defined(HAS_ZFS) */
-
 
 #define	P_ADDR		ki_paddr
 #define	P_COMM		ki_comm
@@ -322,7 +298,15 @@ int     open(const char *, int, ...);
 #define	vasprintf vasprintf_kernel_lsof
 #define	uintfptr_t	int
 #define	_SYS_LIBKERN_H_
+#if __FreeBSD_version > 1400066
+#define	tick_sbt 1
+#define	pause kernel_pause
 #include <sys/file.h>
+#undef	pause
+#undef	tick_sbt
+#else
+#include <sys/file.h>
+#endif
 
 /*
  * Attempt to remove the circumventions.
@@ -343,12 +327,13 @@ int     open(const char *, int, ...);
 #undef	_KERNEL
 #undef	KERNEL
 
-# if	defined(DTYPE_KQUEUE)
 #define	HASKQUEUE				/* has the kqueue file type */
+
+#if __FreeBSD_version < 1400066
 #define	_KERNEL
 #include <sys/eventvar.h>
 #undef	_KERNEL
-# endif	/* defined(DTYPE_KQUEUE) */
+# endif
 
 struct vop_advlock_args { int dummy; };	/* to pacify lf_advlock() prototype */
 #undef	MALLOC_DECLARE
@@ -454,8 +439,7 @@ extern KA_T Kpa;
 # endif	/* defined(P_ADDR) */
 
 struct l_vfs {
-	KA_T addr;			/* kernel address */
-	fsid_t	fsid;			/* file system ID */
+	uint64_t fsid;			/* file system ID */
 
 # if	defined(MOUNT_NONE)
 	short type;			/* type of file system */
@@ -512,6 +496,28 @@ struct sfile {
 #define	XDR_PMAPLIST	(const xdrproc_t)xdr_pmaplist
 
 
+struct pcb_lists {
+	struct xunpcb *un_stream_pcbs;
+	size_t n_un_stream_pcbs;
+	struct xunpcb *un_dgram_pcbs;
+	size_t n_un_dgram_pcbs;
+	struct xunpcb *un_seqpacket_pcbs;
+	size_t n_un_seqpacket_pcbs;
+	struct xtcpcb *tcp_pcbs;
+	size_t n_tcp_pcbs;
+	struct xinpcb *udp_pcbs;
+	size_t n_udp_pcbs;
+};
+
+
+struct lock_list {
+#ifdef KERN_LOCKF
+	struct kinfo_lockf *locks;
+	size_t n_locks;
+#endif
+};
+
+
 /*
  * Definitions for rdev.c
  */
@@ -538,7 +544,7 @@ struct sfile {
  */
 
 struct	namecache {
-#   if  __FreeBSD_version < 1202000 || (__FreeBSD_version >= 1300000 && __FreeBSD_version < 1300105)
+#   if  __FreeBSD_version >= 1300000 && __FreeBSD_version < 1300105
 	LIST_ENTRY(namecache) nc_hash;	/* hash chain */
 	LIST_ENTRY(namecache) nc_src;	/* source vnode list */
 	TAILQ_ENTRY(namecache) nc_dst;	/* destination vnode list */
@@ -584,4 +590,7 @@ struct	namecache {
 #define	VNODE_VFLAG	v_iflag
 #define	NCACHE_VROOT	VV_ROOT
 
+
+#include <libutil.h>
+
 #endif	/* defined(FREEBSD_LSOF_H) */
