@@ -29,12 +29,6 @@
  * 4. This notice may not be removed or altered.
  */
 
-#ifndef lint
-static char copyright[] =
-"@(#) Copyright 1997 Purdue Research Foundation.\nAll rights reserved.\n";
-static char *rcsid = "$Id: dproc.c,v 1.26 2012/04/10 16:39:50 abe Exp $";
-#endif
-
 #include "lsof.h"
 
 
@@ -42,9 +36,32 @@ static char *rcsid = "$Id: dproc.c,v 1.26 2012/04/10 16:39:50 abe Exp $";
  * Local definitions
  */
 
-#define	FDINFO_FLAGS		1	/* fdinfo flags available */
-#define	FDINFO_POS		2	/* fdinfo position available */
-#define FDINFO_ALL		(FDINFO_FLAGS | FDINFO_POS)
+#define	FDINFO_FLAGS		0x1	/* fdinfo flags available */
+#define	FDINFO_POS		0x2	/* fdinfo position available */
+
+#if	defined(HASEPTOPTS)
+#define FDINFO_EVENTFD_ID	0x4	/* fdinfo eventfd-id available */
+#if	defined(HASPTYEPT)
+#define FDINFO_TTY_INDEX	0x8	/* fdinfo tty-index available */
+#endif  /* defined(HASPTYEPT) */
+#endif	/* defined(HASEPTOPTS) */
+
+#define FDINFO_PID		0x10	/* pidfd pid available */
+#define FDINFO_TFD		0x20	/* fd monitored by eventpoll fd */
+
+#define FDINFO_BASE		(FDINFO_FLAGS | FDINFO_POS)
+#if	defined(HASEPTOPTS)
+#if     defined(HASPTYEPT)
+#define FDINFO_ALL		(FDINFO_BASE | FDINFO_TTY_INDEX | FDINFO_EVENTFD_ID | FDINFO_PID | FDINFO_TFD)
+#else  /* !defined(HASPTYEPT) */
+#define FDINFO_ALL		(FDINFO_BASE | FDINFO_EVENTFD_ID | FDINFO_PID | FDINFO_TFD)
+#endif  /* defined(HASPTYEPT) */
+#define FDINFO_OPTIONAL		(FDINFO_ALL & ~FDINFO_BASE)
+#else   /* !defined(HASEPTOPTS) */
+#define FDINFO_ALL		(FDINFO_BASE | FDINFO_PID | FDINFO_TFD)
+#endif	/* defined(HASEPTOPTS) */
+
+
 #define	LSTAT_TEST_FILE		"/"
 #define LSTAT_TEST_SEEK		1
 
@@ -52,6 +69,9 @@ static char *rcsid = "$Id: dproc.c,v 1.26 2012/04/10 16:39:50 abe Exp $";
 #define	ULLONG_MAX		18446744073709551615ULL
 #endif	/* !defined(ULLONG_MAX) */
 
+#define	NS_PATH_LENGTH		100     /* namespace path string length */
+#define	MAP_PATH_LENGTH		100     /* map_files path length */
+#define	ADDR_LENGTH		100     /* addr range of map_files length */
 
 /*
  * Local structures
@@ -60,6 +80,19 @@ static char *rcsid = "$Id: dproc.c,v 1.26 2012/04/10 16:39:50 abe Exp $";
 struct l_fdinfo {
 	int flags;			/* flags: line value */
 	off_t pos;			/* pos: line value */
+
+#if	defined(HASEPTOPTS)
+	int eventfd_id;
+#if	defined(HASPTYEPT)
+	int tty_index;			/* pty line index */
+#endif  /* defined(HASPTYEPT) */
+#endif	/* defined(HASEPTOPTS) */
+
+	int pid;			/* for pidfd */
+
+#define EPOLL_MAX_TFDS 32
+	int tfds [ EPOLL_MAX_TFDS ];
+	size_t tfd_count;
 };
 
 
@@ -80,18 +113,21 @@ static short Ckscko;			/* socket file only checking status:
  * Local function prototypes
  */
 
-_PROTOTYPE(static int get_fdinfo,(char *p, struct l_fdinfo *fi));
-_PROTOTYPE(static int getlinksrc,(char *ln, char *src, int srcl));
+_PROTOTYPE(static MALLOC_S alloc_cbf,(MALLOC_S len, char **cbf, MALLOC_S cbfa));
+_PROTOTYPE(static int get_fdinfo,(char *p, int msk, struct l_fdinfo *fi));
+_PROTOTYPE(static int getlinksrc,(char *ln, char *src, int srcl, char **rest));
 _PROTOTYPE(static int isefsys,(char *path, char *type, int l,
 			       efsys_list_t **rep, struct lfile **lfr));
 _PROTOTYPE(static int nm2id,(char *nm, int *id, int *idl));
-_PROTOTYPE(static int read_id_stat,(int ty, char *p, int id, char **cmd,
-				    int *ppid, int *pgid));
+_PROTOTYPE(static int read_id_stat,(char *p, int id, char **cmd, int *ppid,
+				    int *pgid));
 _PROTOTYPE(static void process_proc_map,(char *p, struct stat *s, int ss));
 _PROTOTYPE(static int process_id,(char *idp, int idpl, char *cmd, UID_ARG uid,
-				  int pid, int ppid, int pgid, int tid));
+				  int pid, int ppid, int pgid, int tid,
+				  char *tcmd));
 _PROTOTYPE(static int statEx,(char *p, struct stat *s, int *ss));
- 
+
+_PROTOTYPE(static void snp_eventpoll, (char *p, int len, int *tfds, int tfd_count));
 
 #if	defined(HASSELINUX)
 _PROTOTYPE(static int cmp_cntx_eq,(char *pcntx, char *ucntx));
@@ -139,7 +175,7 @@ enter_cntx_arg(cntx)
  */
 	if (!(cntxp = (cntxlist_t *)malloc((MALLOC_S)sizeof(cntxlist_t)))) {
 	    (void) fprintf(stderr, "%s: no space for context: %s\n", Pn, cntx);
-	    Exit(1);
+	    Error();
 	}
 	cntxp->f = 0;
 	cntxp->cntx = cntx;
@@ -150,6 +186,29 @@ enter_cntx_arg(cntx)
 #endif	/* defined(HASSELINUX) */
 
 
+/*
+ * alloc_cbf() -- allocate a command buffer
+ */
+
+static MALLOC_S
+alloc_cbf(len, cbf, cbfa)
+	MALLOC_S len;				/* required length */
+	char **cbf;				/* current buffer */
+	MALLOC_S cbfa;				/* current buffer allocation */
+{
+	if (*cbf)
+	    *cbf = (char *)realloc((MALLOC_P *)*cbf, len);
+	else
+	    *cbf = (char *)malloc(len);
+	if (!*cbf) {
+	    (void) fprintf(stderr,
+		"%s: can't allocate command %d bytes\n", Pn, (int)len);
+	     Error();
+	}
+	return(len);
+}
+
+
 /*
  * gather_proc_info() -- gather process information
  */
@@ -158,9 +217,10 @@ void
 gather_proc_info()
 {
 	char *cmd, *tcmd;
+	char cmdbuf[MAXPATHLEN];
 	struct dirent *dp;
 	unsigned char ht, pidts;
-	int n, nl, pgid, pid, ppid, rv, tid, tpgid, tppid, tx;
+	int n, nl, pgid, pid, ppid, prv, rv, tid, tpgid, tppid, tx;
 	static char *path = (char *)NULL;
 	static int pathl = 0;
 	static char *pidpath = (char *)NULL;
@@ -185,7 +245,7 @@ gather_proc_info()
 		(void) fprintf(stderr,
 		    "%s: can't allocate %d bytes for \"%s/\"<pid>\n",
 		    Pn, (int)pidpathl, PROCFS);
-		Exit(1);
+		Error();
 	    }
 	    (void) snpf(pidpath, pidpathl, "%s/", PROCFS);
 	}
@@ -228,7 +288,7 @@ gather_proc_info()
 	     * If only ORed process selection options have been specified,
 	     * enable conditional file skipping and socket file only checking.
 	     */
-		if ((Selflags & SELFILE) || !(Selflags & SELPROC))
+		if ((Selflags & SELFILE) || !(Selflags & SelProc))
 		    Cckreg = Ckscko = 0;
 		else
 		    Cckreg = Ckscko = 1;
@@ -248,7 +308,7 @@ gather_proc_info()
 	if (!ps) {
 	    if (!(ps = opendir(PROCFS))) {
 		(void) fprintf(stderr, "%s: can't open %s\n", Pn, PROCFS);
-		Exit(1);
+		Error();
 	    }
 	} else
 	    (void) rewinddir(ps);
@@ -265,7 +325,7 @@ gather_proc_info()
 		    (void) fprintf(stderr,
 			"%s: can't allocate %d bytes for \"%s/%s/\"\n",
 			Pn, (int)pidpathl, PROCFS, dp->d_name);
-		    Exit(1);
+		    Error();
 		}
 	    }
 	    (void) snpf(pidpath + pidx, pidpathl - pidx, "%s/", dp->d_name);
@@ -277,13 +337,26 @@ gather_proc_info()
 		continue;
 	    uid = (UID_ARG)sb.st_uid;
 	    ht = pidts = 0;
+	/*
+	 * Get the PID's command name.
+	 */
+	    (void) make_proc_path(pidpath, n, &path, &pathl, "stat");
+	    if ((prv = read_id_stat(path, pid, &cmd, &ppid, &pgid)) < 0)
+		cmd = "(unknown)";
 
 #if	defined(HASTASKS)
 	/*
-	 * If task reporting is selected, check the tasks of the process first,
-	 * so that the "-p<PID> -aK" options work properly.
+	 * Task reporting has been selected, so save the process' command
+	 * string, so that task processing won't change it in the buffer of
+	 * read_id_stat().
+	 *
+	 * Check the tasks of the process first, so that the "-p<PID> -aK"
+	 * options work properly.
 	 */
-	    if ((Selflags & SELTASK)) {
+	    else if (!IgnTasks && (Selflags & SELTASK)) {
+		strncpy(cmdbuf, cmd, sizeof(cmdbuf) - 1);
+		cmdbuf[sizeof(cmdbuf) - 1] = '\0';
+		cmd = cmdbuf;
 		(void) make_proc_path(pidpath, n, &taskpath, &taskpathl,
 				      "task");
 		tx = n + 4;
@@ -322,7 +395,7 @@ gather_proc_info()
 				    tidpathl);
 				(void) fprintf(stderr, " for \"%s/%s/stat\"\n",
 				    taskpath, dp->d_name);
-				Exit(1);
+				Error();
 			    }
 			}
 			(void) snpf(tidpath, tidpathl, "%s/%s/stat", taskpath,
@@ -330,15 +403,15 @@ gather_proc_info()
 		    /*
 		     * Check the task state.
 		     */
-			rv = read_id_stat(1, tidpath, tid, &tcmd, &tppid,
+			rv = read_id_stat(tidpath, tid, &tcmd, &tppid,
 					  &tpgid);
 			if ((rv < 0) || (rv == 1))
 			    continue;
 		    /*
 		     * Attempt to record the task.
 		     */
-			if (!process_id(tidpath, (tx + 1 + nl+ 1), tcmd, uid,
-					pid, tppid, tpgid, tid))
+			if (!process_id(tidpath, (tx + 1 + nl+ 1), cmd, uid,
+					pid, tppid, tpgid, tid, tcmd))
 			{
 			    ht = 1;
 			}
@@ -351,15 +424,14 @@ gather_proc_info()
 	/*
 	 * If the main process is a task and task selection has been specified
 	 * along with option ANDing, enter the main process temporarily as a
-	 * task,  so that the "-aK" option set lists the main process along
+	 * task, so that the "-aK" option set lists the main process along
 	 * with its tasks.
 	 */
-	    (void) make_proc_path(pidpath, n, &path, &pathl, "stat");
-	    if (((rv = read_id_stat(0, path, pid, &cmd, &ppid, &pgid)) >= 0) 
-	    &&   (rv != 1))
-	    {
-		tid = (Fand && ht && pidts && (Selflags & SELTASK)) ? pid : 0;
-		if ((!process_id(pidpath, n, cmd, uid, pid, ppid, pgid, tid))
+	    if ((prv >= 0) && (prv != 1)) {
+		tid = (Fand && ht && pidts && !IgnTasks && (Selflags & SELTASK))
+		    ? pid : 0;
+		if ((!process_id(pidpath, n, cmd, uid, pid, ppid, pgid, tid,
+				 (char *)NULL))
 		&&  tid)
 		{
 		    Lp->tid = 0;
@@ -374,8 +446,10 @@ gather_proc_info()
  */
 
 static int
-get_fdinfo(p, fi)
+get_fdinfo(p, msk, fi)
 	char *p;			/* path to fdinfo file */
+	int msk;			/* mask for information type: e.g.,
+					 * the FDINFO_* definition */
 	struct l_fdinfo *fi;		/* pointer to local fdinfo values
 					 * return structure */
 {
@@ -390,17 +464,28 @@ get_fdinfo(p, fi)
  */
 	if (!fi)
 	    return(0);
+
+#if	defined(HASEPTOPTS)
+	fi->eventfd_id = -1;
+#if	defined(HASPTYEPT)
+	fi->tty_index = -1;
+#endif	/* defined(HASPTYEPT) */
+#endif	/* defined(HASEPTOPTS) */
+	fi->pid = -1;
+	fi->tfd_count = 0;
+
 	if (!p || !*p || !(fs = fopen(p, "r")))
 	    return(0);
 /*
  * Read the fdinfo file.
  */
 	while (fgets(buf, sizeof(buf), fs)) {
+	    int opt_flg = 0;
 	    if (get_fields(buf, (char *)NULL, &fp, (int *)NULL, 0) < 2)
 		continue;
 	    if (!fp[0] || !*fp[0] || !fp[1] || !*fp[1])
 		continue;
-	    if (!strcmp(fp[0], "flags:")) {
+	    if ((msk & FDINFO_FLAGS) && !strcmp(fp[0], "flags:")) {
 
 	    /*
 	     * Process a "flags:" line.
@@ -410,9 +495,9 @@ get_fdinfo(p, fi)
 		||  !ep || *ep)
 		    continue;
 		fi->flags = (unsigned int)ul;
-		if ((rv |= FDINFO_FLAGS) == FDINFO_ALL)
+		if ((rv |= FDINFO_FLAGS) == msk)
 		    break;
-	    } else if (!strcmp(fp[0], "pos:")) {
+	    } else if ((msk & FDINFO_POS) && !strcmp(fp[0], "pos:")) {
 
 	    /*
 	     * Process a "pos:" line.
@@ -422,8 +507,76 @@ get_fdinfo(p, fi)
 		||  !ep || *ep)
 		    continue;
 		fi->pos = (off_t)ull;
-		if ((rv |= FDINFO_POS) == FDINFO_ALL)
+		if ((rv |= FDINFO_POS) == msk)
 		    break;
+
+	    } else if (
+		       ((msk & FDINFO_PID) && !strcmp(fp[0], "Pid:")
+			&& ((opt_flg = FDINFO_PID)))
+		       || ((msk & FDINFO_TFD) && !strcmp(fp[0], "tfd:")
+			   && ((opt_flg = FDINFO_TFD)))
+#if	defined(HASEPTOPTS)
+		       || ((msk & FDINFO_EVENTFD_ID) && !strcmp(fp[0], "eventfd-id:")
+			   && ((opt_flg = FDINFO_EVENTFD_ID)))
+#if	defined(HASPTYEPT)
+		       || ((msk & FDINFO_TTY_INDEX) && !strcmp(fp[0], "tty-index:")
+			   && ((opt_flg = FDINFO_TTY_INDEX)))
+#endif	/* defined(HASPTYEPT) */
+#endif	/* defined(HASEPTOPTS) */
+		       ) {
+		int val;
+	    /*
+	     * Process a "tty-index:", "eventfd-id:", "Pid:", or "tfid:" line.
+	     */
+		ep = (char *)NULL;
+		if ((ul = strtoul(fp[1], &ep, 0)) == ULONG_MAX
+		||  !ep || *ep)
+		     continue;
+
+		val = (int)ul;
+		if (val < 0) {
+		/*
+		 * Oops! If integer overflow occurred, reset the field.
+		 */
+		     val = -1;
+		}
+
+		rv |= opt_flg;
+		switch (opt_flg) {
+#if	defined(HASEPTOPTS)
+		case FDINFO_EVENTFD_ID:
+		    fi->eventfd_id = val;
+		    break;
+#if	defined(HASPTYEPT)
+		case FDINFO_TTY_INDEX:
+		    fi->tty_index = val;
+		    break;
+#endif	/* defined(HASPTYEPT) */
+#endif	/* defined(HASEPTOPTS) */
+		case FDINFO_PID:
+		    fi->pid = val;
+		    break;
+		case FDINFO_TFD:
+		    if (fi->tfd_count < EPOLL_MAX_TFDS) {
+			fi->tfds [fi->tfd_count] = val;
+			fi->tfd_count++;
+		    }
+		    break;
+		}
+
+		if ((
+			/* There can be more than one tfd: lines.
+			   So even if we found one, we can not exit the loop.
+			   However, we can assume tfd lines are continuous. */
+			opt_flg != FDINFO_TFD
+			&& (rv == msk || (rv & FDINFO_TFD))
+		    )
+		    || (
+			/* Too many tfds. */
+			opt_flg == FDINFO_TFD
+			&& rv == msk && fi->tfd_count == EPOLL_MAX_TFDS
+		    ))
+		  break;
 	    }
 	}
 	fclose(fs);
@@ -440,14 +593,19 @@ get_fdinfo(p, fi)
 
 
 static int
-getlinksrc(ln, src, srcl)
+getlinksrc(ln, src, srcl, rest)
 	char *ln;			/* link path */
 	char *src;			/* link source path return address */
 	int srcl;			/* length of src[] */
+	char **rest;			/* pointer to what follows the ':' in
+					 * the link source path (NULL if no
+					 * return requested) */
 {
 	char *cp;
 	int ll;
 
+	if (rest)
+	    *rest = (char *)NULL;
 	if ((ll = readlink(ln, src, srcl - 1)) < 1
 	||  ll >= srcl)
 	    return(-1);
@@ -457,6 +615,8 @@ getlinksrc(ln, src, srcl)
 	if ((cp = strchr(src, ':'))) {
 	    *cp = '\0';
 	    ll = strlen(src);
+	    if (rest)
+		*rest = cp + 1;
 	}
 	return(ll);
 }
@@ -503,7 +663,7 @@ initialize()
 	    if (!OffType) {
 		(void) snpf(path, sizeof(path), "%s/%d/fdinfo/%d", PROCFS,
 			    Mypid, fd);
-		if (get_fdinfo(path, &fi) & FDINFO_POS) {
+		if (get_fdinfo(path, FDINFO_POS, &fi) & FDINFO_POS) {
 		    if (fi.pos == (off_t)LSTAT_TEST_SEEK)
 			OffType = 2;
 		}
@@ -571,7 +731,7 @@ make_proc_path(pp, pl, np, nl, sf)
 		(void) fprintf(stderr,
 		    "%s: can't allocate %d bytes for %s%s\n",
 		    Pn, (int)rl, pp, sf);
-		Exit(1);
+		Error();
 	    }
 	    *nl = rl;
 	    *np = cp;
@@ -705,7 +865,7 @@ open_proc_stream(p, m, buf, sz, act)
 	int act;			/* fopen() failure action:
 					 *     0 : return (FILE *)NULL
 					 *   <>0 : fprintf() an error message
-					 *         and Exit(1)
+					 *         and Error()
 					 */
 {
 	FILE *fs;			/* opened stream */
@@ -719,7 +879,7 @@ open_proc_stream(p, m, buf, sz, act)
 		return((FILE *)NULL);
 	    (void) fprintf(stderr, "%s: can't fopen(%s, \"%s\"): %s\n",
 		Pn, p, m, strerror(errno));
-	    Exit(1);
+	    Error();
 	}
 /*
  * Return the stream if no buffer change is required.
@@ -742,7 +902,7 @@ open_proc_stream(p, m, buf, sz, act)
 		(void) fprintf(stderr,
 		    "%s: can't allocate %d bytes for %s stream buffer\n",
 		    Pn, (int)tsz, p);
-		Exit(1);
+		Error();
 	    }
 	    *sz = tsz;
 	}
@@ -752,7 +912,7 @@ open_proc_stream(p, m, buf, sz, act)
 	if (setvbuf(fs, *buf, _IOFBF, tsz)) {
 	    (void) fprintf(stderr, "%s: setvbuf(%s)=%d failure: %s\n",
 		Pn, p, (int)tsz, strerror(errno));
-	    Exit(1);
+	    Error();
 	}
 	return(fs);
 }
@@ -766,7 +926,7 @@ open_proc_stream(p, m, buf, sz, act)
  */
 
 static int
-process_id(idp, idpl, cmd, uid, pid, ppid, pgid, tid)
+process_id(idp, idpl, cmd, uid, pid, ppid, pgid, tid, tcmd)
 	char *idp;			/* pointer to ID's path */
 	int idpl;			/* pointer to ID's path length */
 	char *cmd;			/* pointer to ID's command */
@@ -775,8 +935,9 @@ process_id(idp, idpl, cmd, uid, pid, ppid, pgid, tid)
 	int ppid;			/* parent PID */
 	int pgid;			/* parent GID */
 	int tid;			/* task ID, if non-zero */
+	char *tcmd;			/* task command, if non-NULL) */
 {
-	int av;
+	int av = 0;
 	static char *dpath = (char *)NULL;
 	static int dpathl = 0;
 	short efs, enls, enss, lnk, oty, pn, pss, sf;
@@ -794,6 +955,7 @@ process_id(idp, idpl, cmd, uid, pid, ppid, pgid, tid)
 	static int pathl = 0;
 	static char *pathi = (char *)NULL;
 	static int pathil = 0;
+	char *rest;
 	int txts = 0;
 
 #if	defined(HASSELINUX)
@@ -805,29 +967,54 @@ process_id(idp, idpl, cmd, uid, pid, ppid, pgid, tid)
  */
 	if (is_proc_excl(pid, pgid, uid, &pss, &sf, tid)
 	||  is_cmd_excl(cmd, &pss, &sf))
+	{
+
+#if	defined(HASEPTOPTS)
+	    if (!FeptE)
+		return(1);
+#else	/* !defined(HASEPTOPTS) */
 	    return(1);
-	if (Cckreg) {
+#endif	/* defined(HASEPTOPTS) */
+
+	}
+	if (Cckreg && !FeptE) {
 
 	/*
 	 * If conditional checking of regular files is enabled, enable
 	 * socket file only checking, based on the process' selection
 	 * status.
 	 */
-	    Ckscko = (sf & SELPROC) ? 0 : 1;
+	    Ckscko = (sf & SelProc) ? 0 : 1;
 	}
 	alloc_lproc(pid, pgid, ppid, uid, cmd, (int)pss, (int)sf);
-	Lp->tid = tid;
 	Plf = (struct lfile *)NULL;
+
+#if	defined(HASTASKS)
+/*
+ * Enter task information.
+ */
+	Lp->tid = tid;
+	if (tid && tcmd) {
+	    if (!(Lp->tcmd = mkstrcpy(tcmd, (MALLOC_S *)NULL))) {
+		(void) fprintf(stderr,
+		    "%s: PID %d, TID %d, no space for task name: ",
+		    Pn, pid, tid);
+		safestrprt(tcmd, stderr, 1);
+		Error();
+	    }
+	}
+#endif	/* defined(HASTASKS) */
+
 /*
  * Process the ID's current working directory info.
  */
+	efs = 0;
 	if (!Ckscko) {
 	    (void) make_proc_path(idp, idpl, &path, &pathl, "cwd");
 	    alloc_lfile(CWD, -1);
-	    efs = 0;
-	    if (getlinksrc(path, pbuf, sizeof(pbuf)) < 1) {
+	    if (getlinksrc(path, pbuf, sizeof(pbuf), (char **)NULL) < 1) {
 		if (!Fwarn) {
-		    (void) memset((void *)&sb, 0, sizeof(sb));
+		    zeromem((char *)&sb, sizeof(sb));
 		    lnk = ss = 0;
 		    (void) snpf(nmabuf, sizeof(nmabuf), "(readlink: %s)",
 			strerror(errno));
@@ -861,7 +1048,7 @@ process_id(idp, idpl, cmd, uid, pid, ppid, pgid, tid)
 	    }
 	    if (pn) {
 		(void) process_proc_node(lnk ? pbuf : path,
-					 &sb, ss,
+					 path, &sb, ss,
 					 (struct stat *)NULL, 0);
 		if (Lf->sf)
 		    link_lfile();
@@ -870,13 +1057,13 @@ process_id(idp, idpl, cmd, uid, pid, ppid, pgid, tid)
 /*
  * Process the ID's root directory info.
  */
+	lnk = ss = 0;
 	if (!Ckscko) {
 	    (void) make_proc_path(idp, idpl, &path, &pathl, "root");
 	    alloc_lfile(RTD, -1);
-	    if (getlinksrc(path, pbuf, sizeof(pbuf)) < 1) {
+	    if (getlinksrc(path, pbuf, sizeof(pbuf), (char **)NULL) < 1) {
 		if (!Fwarn) {
-		    (void) memset((void *)&sb, 0, sizeof(sb));
-		    lnk = ss = 0;
+		    zeromem((char *)&sb, sizeof(sb));
 		    (void) snpf(nmabuf, sizeof(nmabuf), "(readlink: %s)",
 			strerror(errno));
 		    nmabuf[sizeof(nmabuf) - 1] = '\0';
@@ -908,7 +1095,7 @@ process_id(idp, idpl, cmd, uid, pid, ppid, pgid, tid)
 	    }
 	    if (pn) {
 		(void) process_proc_node(lnk ? pbuf : path,
-					 &sb, ss,
+					 path, &sb, ss,
 					 (struct stat *)NULL, 0);
 		if (Lf->sf)
 		    link_lfile();
@@ -917,13 +1104,12 @@ process_id(idp, idpl, cmd, uid, pid, ppid, pgid, tid)
 /*
  * Process the ID's execution info.
  */
+	lnk = ss = txts = 0;
 	if (!Ckscko) {
-	    txts = 0;
 	    (void) make_proc_path(idp, idpl, &path, &pathl, "exe");
 	    alloc_lfile("txt", -1);
-	    if (getlinksrc(path, pbuf, sizeof(pbuf)) < 1) {
-		(void) memset((void *)&sb, 0, sizeof(sb));
-		lnk = ss = 0;
+	    if (getlinksrc(path, pbuf, sizeof(pbuf), (char **)NULL) < 1) {
+		zeromem((void *)&sb, sizeof(sb));
 		if (!Fwarn) {
 		    if ((errno != ENOENT) || uid) {
 			(void) snpf(nmabuf, sizeof(nmabuf), "(readlink: %s)",
@@ -962,7 +1148,7 @@ process_id(idp, idpl, cmd, uid, pid, ppid, pgid, tid)
 	    }
 	    if (pn) {
 		(void) process_proc_node(lnk ? pbuf : path,
-					 &sb, ss,
+					 path, &sb, ss,
 					 (struct stat *)NULL, 0);
 		if (Lf->sf)
 		    link_lfile();
@@ -996,7 +1182,7 @@ process_id(idp, idpl, cmd, uid, pid, ppid, pgid, tid)
 			(void) fprintf(stderr,
 			    "%s: no context error space: PID %ld",
 			    Pn, (long)Lp->pid);
-			Exit(1);
+			Error();
 		    }
 		}
 	    } else if (CntxArg) {
@@ -1044,10 +1230,11 @@ process_id(idp, idpl, cmd, uid, pid, ppid, pgid, tid)
 		continue;
 	    (void) make_proc_path(dpath, i, &path, &pathl, fp->d_name);
 	    (void) alloc_lfile((char *)NULL, fd);
-	    if (getlinksrc(path, pbuf, sizeof(pbuf)) < 1) {
-		(void) memset((void *)&sb, 0, sizeof(sb));
+	    if (getlinksrc(path, pbuf, sizeof(pbuf), &rest) < 1) {
+		zeromem((char *)&sb, sizeof(sb));
 		lnk = ss = 0;
 		if (!Fwarn) {
+		    ls = 0;
 		    (void) snpf(nmabuf, sizeof(nmabuf), "(readlink: %s)",
 			strerror(errno));
 		    nmabuf[sizeof(nmabuf) - 1] = '\0';
@@ -1107,9 +1294,29 @@ process_id(idp, idpl, cmd, uid, pid, ppid, pgid, tid)
 	    }
 	    if (pn || (efs && lfr && oty)) {
 		if (oty) {
+		    int fdinfo_mask = FDINFO_BASE;
 		    (void) make_proc_path(ipath, j, &pathi, &pathil,
 					  fp->d_name);
-		    if ((av = get_fdinfo(pathi, &fi)) & FDINFO_POS) {
+
+		    if (rest && rest[0] == '['
+			&& rest[1] == 'e' && rest[2] == 'v' && rest[3] == 'e'
+			&& rest[4] == 'n' && rest[5] == 't') {
+#if	defined(HASEPTOPTS)
+			if (rest[6] == 'f')
+			    fdinfo_mask |= FDINFO_EVENTFD_ID;
+#endif	/* defined(HASEPTOPTS) */
+			else if (rest[6] == 'p')
+			    fdinfo_mask |= FDINFO_TFD;
+		    }
+#if	defined(HASEPTOPTS)
+#if	defined(HASPTYEPT)
+		    fdinfo_mask |= FDINFO_TTY_INDEX;
+#endif  /* defined(HASPTYEPT) */
+#endif	/* defined(HASEPTOPTS) */
+		    if (rest && rest[0] == '[' && rest[1] == 'p')
+			fdinfo_mask |= FDINFO_PID;
+
+		    if ((av = get_fdinfo(pathi,fdinfo_mask,&fi)) & FDINFO_POS) {
 			if (efs) {
 			    if (Foffset) {
 				lfr->off = (SZOFFTYPE)fi.pos;
@@ -1136,7 +1343,52 @@ process_id(idp, idpl, cmd, uid, pid, ppid, pgid, tid)
 
 		}
 		if (pn) {
-		    process_proc_node(lnk ? pbuf : path, &sb, ss, &lsb, ls);
+		    process_proc_node(lnk ? pbuf : path, path, &sb, ss, &lsb,
+				      ls);
+		    if (Lf->ntype == N_ANON_INODE) {
+			if (rest && *rest) {
+#if	defined(HASEPTOPTS)
+			    if (fi.eventfd_id != -1
+				&& strcmp(rest, "[eventfd]") == 0) {
+				(void) snpf(rest,
+					    sizeof(pbuf) - (rest - pbuf),
+					    "[eventfd:%d]", fi.eventfd_id);
+			    }
+#endif	/* defined(HASPTYEPT) */
+			    if (fi.pid != -1
+				&& strcmp(rest, "[pidfd]") == 0) {
+				(void) snpf (rest,
+					     sizeof(pbuf) - (rest - pbuf),
+					     "[pidfd:%d]", fi.pid);
+			    }
+			    if (fi.tfd_count > 0
+				&& strcmp(rest, "[eventpoll]") == 0) {
+				snp_eventpoll (rest, sizeof(pbuf) - (rest - pbuf),
+					       fi.tfds, fi.tfd_count);
+			    }
+
+			    enter_nm(rest);
+			}
+#if	defined(HASEPTOPTS)
+			if (FeptE && fi.eventfd_id != -1) {
+			    enter_evtfdinfo(fi.eventfd_id);
+			    Lf->eventfd_id = fi.eventfd_id;
+			    Lf->sf |= SELEVTFDINFO;
+			}
+#endif	/* defined(HASPTYEPT) */
+		    }
+#if	defined(HASEPTOPTS) && defined(HASPTYEPT)
+		    else if (FeptE
+			 &&  Lf->rdev_def
+			 &&  is_pty_ptmx(Lf->rdev)
+			 &&  (av & FDINFO_TTY_INDEX)
+		    ) {
+			    enter_ptmxi(fi.tty_index);
+			    Lf->tty_index = fi.tty_index;
+			    Lf->sf |= SELPTYINFO;
+		    }
+#endif	/* defined(HASEPTOPTS) && defined(HASPTYEPT) */
+
 		    if (Lf->sf)
 			link_lfile();
 		}
@@ -1146,6 +1398,31 @@ process_id(idp, idpl, cmd, uid, pid, ppid, pgid, tid)
 	return(0);
 }
 
+/* compare mount namespace of this lsof process and the target process */
+
+static int
+compare_mntns(pid)
+        int pid;                        /* pid of the target process */
+{
+	char nspath[NS_PATH_LENGTH];
+	struct stat sb_self, sb_target;
+	int ret;
+
+	if (stat("/proc/self/ns/mnt", &sb_self))
+	    return -1;
+
+	ret = snprintf(nspath, sizeof(nspath), "/proc/%d/ns/mnt", pid);
+	if (ret >= sizeof(nspath) || ret <= 0)
+	    return -1;
+
+	if (stat(nspath, &sb_target))
+	    return -1;
+
+	if (sb_self.st_ino != sb_target.st_ino)
+	    return -1;
+
+	return 0;
+}
 
 /*
  * process_proc_map() - process the memory map of a process
@@ -1176,12 +1453,18 @@ process_proc_map(p, s, ss)
 	static int sma = 0;
 	static char *vbuf = (char *)NULL;
 	static size_t vsz = (size_t)0;
+	int diff_mntns = 0;
 /*
  * Open the /proc/<pid>/maps file, assign a page size buffer to its stream,
  * and read it/
  */
 	if (!(ms = open_proc_stream(p, "r", &vbuf, &vsz, 0)))
 	    return;
+
+	/* target process in a different mount namespace from lsof process. */
+	if (compare_mntns(Lp->pid))
+	    diff_mntns = 1;
+
 	while (fgets(buf, sizeof(buf), ms)) {
 	    if ((nf = get_fields(buf, ":", &fp, &eb, 1)) < 7)
 		continue;			/* not enough fields */
@@ -1254,7 +1537,7 @@ process_proc_map(p, s, ss)
 		    (void) fprintf(stderr,
 			"%s: can't allocate %d bytes for saved maps, PID %d\n",
 			Pn, (int)len, Lp->pid);
-		    Exit(1);
+		    Error();
 		}
 	    }
 	    sm[ns].dev = dev;
@@ -1269,11 +1552,44 @@ process_proc_map(p, s, ss)
 		efs = sv = 1;
 	    else
 		efs = 0;
-	    if (!efs) {
-		if (HasNFS)
-		    sv = statsafely(fp[6], &sb);
-		else
-		    sv = stat(fp[6], &sb);
+
+	    /* For processes in different mount namespace from lsof process,
+	     * stat corresponding files under /proc/[pid]/map_files would follow
+	     * symlinks regardless of namespaces.
+	     */
+	    if (diff_mntns) {
+		char path[MAP_PATH_LENGTH];
+		char addr[ADDR_LENGTH];
+		uint64_t start, end;
+		int ret;
+
+		if (sscanf(fp[0], "%lx-%lx", &start, &end) != 2)
+		    goto stat_directly;
+
+		ret = snprintf(addr, sizeof(addr), "%lx-%lx", start, end);
+		if (ret >= sizeof(addr) || ret <= 0)
+		    goto stat_directly;
+
+		ret = snprintf(path, sizeof(path), "/proc/%d/map_files/%s",
+			       Lp->pid, addr);
+		if (ret >= sizeof(path) || ret <= 0)
+		    goto stat_directly;
+
+		if (!efs) {
+		    if (HasNFS)
+		        sv = statsafely(path, &sb);
+		    else
+		        sv = stat(path, &sb);
+	        }
+	    }
+	    else {
+stat_directly:
+		if (!efs) {
+		    if (HasNFS)
+			sv = statsafely(fp[6], &sb);
+		    else
+			sv = stat(fp[6], &sb);
+		}
 	    }
 	    if (sv || efs) {
 		en = errno;
@@ -1282,10 +1598,10 @@ process_proc_map(p, s, ss)
 	     * exempt file system) or stat(2) failed, so manufacture a partial
 	     * stat(2) reply from the process' maps file entry.
 	     *
-	     * If the file has been deleted, reset its type to "DEL"; otherwise
-	     * generate a stat() error name addition.
+	     * If the file has been deleted, reset its type to "DEL";
+	     * otherwise generate a stat() error name addition.
 	     */
-		(void) memset((void *)&sb, 0, sizeof(sb));
+		zeromem((char *)&sb, sizeof(sb));
 		sb.st_dev = dev;
 		sb.st_ino = (ino_t)inode;
 		sb.st_mode = S_IFREG;
@@ -1298,6 +1614,8 @@ process_proc_map(p, s, ss)
 		    nmabuf[sizeof(nmabuf) - 1] = '\0';
 		    (void) add_nma(nmabuf, strlen(nmabuf));
 		}
+	    } else if (diff_mntns) {
+		mss = SB_ALL;
 	    } else if ((sb.st_dev != dev) || ((INODETYPE)sb.st_ino != inode)) {
 
 	    /*
@@ -1334,7 +1652,7 @@ process_proc_map(p, s, ss)
 			(void) add_nma(nmabuf, strlen(nmabuf));
 		    }
 		}
-		(void) memset((void *)&sb, 0, sizeof(sb));
+		zeromem((char *)&sb, sizeof(sb));
 		sb.st_dev = dev;
 		sb.st_ino = (ino_t)inode;
 		sb.st_mode = S_IFREG;
@@ -1345,7 +1663,8 @@ process_proc_map(p, s, ss)
 	 * Record the file's information.
 	 */
 	    if (!efs)
-		process_proc_node(fp[6], &sb, mss, (struct stat *)NULL, 0);
+		process_proc_node(fp[6], fp[6], &sb, mss, (struct stat *)NULL,
+				  0);
 	    else {
 
 	    /*
@@ -1381,8 +1700,7 @@ process_proc_map(p, s, ss)
  */
 
 static int
-read_id_stat(ty, p, id, cmd, ppid, pgid)
-	int ty;				/* type: 0 == PID, 1 == LWP */
+read_id_stat(p, id, cmd, ppid, pgid)
 	char *p;			/* path to status file */
 	int id;				/* ID: PID or LWP */
 	char **cmd;			/* malloc'd command name */
@@ -1391,11 +1709,10 @@ read_id_stat(ty, p, id, cmd, ppid, pgid)
 					 * type */
 {
 	char buf[MAXPATHLEN], *cp, *cp1, **fp;
+	int ch, cx, es, nf, pc;
 	static char *cbf = (char *)NULL;
 	static MALLOC_S cbfa = 0;
 	FILE *fs;
-	MALLOC_S len;
-	int nf;
 	static char *vbuf = (char *)NULL;
 	static size_t vsz = (size_t)0;
 /*
@@ -1404,69 +1721,102 @@ read_id_stat(ty, p, id, cmd, ppid, pgid)
  */
 	if (!(fs = open_proc_stream(p, "r", &vbuf, &vsz, 0)))
 	    return(-1);
-	cp = fgets(buf, sizeof(buf), fs);
-	(void) fclose(fs);
-	if (!cp)
-	    return(-1);
-/*
- * Separate the line into fields on white space separators.  Expect five fields
- * for a PID type and three for an LWP type.
- */
-	if ((nf = get_fields(buf, (char *)NULL, &fp, (int *)NULL, 0))
-	<  (ty ? 5 : 3))
-	{
+	if (!(cp = fgets(buf, sizeof(buf), fs))) {
+
+read_id_stat_exit:
+
+	    (void) fclose(fs);
 	    return(-1);
 	}
 /*
- * Convert the first field to an integer; its conversion must match the
- * ID argument.
+ * Skip to the first field, and make sure it is a matching ID.
  */
-	if (!fp[0] || (atoi(fp[0]) != id))
-	    return(-1);
-/*
- * Get the command name from the second field.  Strip a starting '(' and
- * an ending ')'.  Allocate space to hold the result and return the space
- * pointer.
- */
-	if (!(cp = fp[1]))
-	    return(-1);
-	if (cp && *cp == '(')
+	cp1 = cp;
+	while (*cp && (*cp != ' ') && (*cp != '\t'))
 	    cp++;
-	if ((cp1 = strrchr(cp, ')')))
-	    *cp1 = '\0';
-	if ((len = strlen(cp) + 1) > cbfa) {
-	     cbfa = len;
-	     if (cbf)
-		cbf = (char *)realloc((MALLOC_P *)cbf, cbfa);
-	     else
-		cbf = (char *)malloc(cbfa);
-	     if (!cbf) {
-		(void) fprintf(stderr,
-		    "%s: can't allocate %d bytes for command \"%s\"\n",
-		    Pn, (int)cbfa, cp);
-		Exit(1);
-	     }
+	if (*cp)
+	    *cp = '\0';
+	if (atoi(cp1) != id)
+	    goto read_id_stat_exit;
+/*
+ * The second field should contain the command, enclosed in parentheses.
+ * If it also has embedded '\n' characters, replace them with '?' characters,
+ * accumulating command characters until a closing parentheses appears.
+ *
+ */
+	for (++cp; *cp && (*cp == ' '); cp++)
+		;
+	if (!cp || (*cp != '('))
+	    goto read_id_stat_exit;
+	cp++;
+	pc = 1;			/* start the parenthesis balance count at 1 */
+/*
+ * Enter the command characters safely.  Supply them from the initial read
+ * of the stat file line, a '\n' if the initial read didn't yield a ')'
+ * command closure, or by reading the rest of the command a character at
+ * a time from the stat file.  Count embedded '(' characters and balance
+ * them with embedded ')' characters.  The opening '(' starts the balance
+ * count at one.
+ */
+	for (cx = es = 0;;) {
+	    if (!es)
+		ch = *cp++;
+	    else {
+		if ((ch = fgetc(fs)) == EOF)
+		    goto read_id_stat_exit;
+	    }
+	    if (ch == '(')		/* a '(' advances the balance count */
+		pc++;
+	    if (ch == ')') {
+
+	    /*
+	     * Balance parentheses when a closure is encountered.  When
+	     * they are balanced, this is the end of the command.
+	     */
+		pc--;
+		if (!pc)
+		    break;
+	    }
+	    if ((cx + 2) > cbfa)
+		cbfa = alloc_cbf((cx + 2), &cbf, cbfa);
+	    cbf[cx] = ch;
+	    cx++;
+	    cbf[cx] = '\0';
+	    if (!es && !*cp)
+		es = 1;		/* Switch to fgetc() when a '\0' appears. */
 	}
-	(void) snpf(cbf, len, "%s", cp);
 	*cmd = cbf;
+/*
+ * Read the remainder of the stat line if it was necessary to read command
+ * characters individually from the stat file.
+ *
+ * Separate the reminder into fields.
+ */
+	if (es)
+	    cp = fgets(buf, sizeof(buf), fs);
+	(void) fclose(fs);
+	if (!cp || !*cp)
+	    return(-1);
+	if ((nf = get_fields(cp, (char *)NULL, &fp, (int *)NULL, 0)) < 3)
+	    return(-1);
 /*
  * Convert and return parent process (fourth field) and process group (fifth
  * field) IDs.
  */
-	if (fp[3] && *fp[3])
-	    *ppid = atoi(fp[3]);
+	if (fp[1] && *fp[1])
+	    *ppid = atoi(fp[1]);
 	else
 	    return(-1);
-	if (fp[4] && *fp[4])
-	    *pgid = atoi(fp[4]);
+	if (fp[2] && *fp[2])
+	    *pgid = atoi(fp[2]);
 	else
 	    return(-1);
 /*
  * Check the state in the third field.  If it is 'Z', return that indication.
  */
-	if (fp[2] && !strcmp(fp[2], "Z"))
+	if (fp[0] && !strcmp(fp[0], "Z"))
 	    return(1);
-	else if (fp[2] && !strcmp(fp[2], "T"))
+	else if (fp[0] && !strcmp(fp[0], "T"))
 	    return(2);
 	return(0);
 }
@@ -1512,7 +1862,7 @@ statEx(p, s, ss)
 		(void) fprintf(stderr,
 		    "%s: PID %ld: no statEx path space: %s\n",
 		    Pn, (long)Lp->pid, p);
-		Exit(1);
+		Error();
 	    }
 	    ca = sz + 1;
 	}
@@ -1537,7 +1887,7 @@ statEx(p, s, ss)
  * If a stat() on a trimmed result succeeded, form partial results containing
  * only the device and raw device numbers.
  */
-	memset((void *)s, 0, sizeof(struct stat));
+	zeromem((char *)s, sizeof(struct stat));
 	if (st) {
 	    errno = 0;
 	    s->st_dev = sb.st_dev;
@@ -1549,3 +1899,85 @@ statEx(p, s, ss)
 	*ss = 0;
 	return(1);
 }
+
+static int
+cal_order (int n)
+{
+	int i = 0;
+	do {
+	    n /= 10;
+	    i++;
+	} while (n);
+	return i;
+}
+
+static int
+snp_eventpoll_fds(char *p, int len, int *tfds, int count)
+{
+	int wl = 0;
+	int i;
+
+	if (len > 0)
+	    p[0] = '\0';
+
+	for (i = 0; i < count; i++) {
+	    int is_last_item = (i == count - 1);
+	    int needs = cal_order (tfds [i]) + (is_last_item? 0: 3);
+
+	    if ((len - wl + (wl? 2: 0)) <= needs) {
+		/* No space to print the tfd. */
+		break;
+	    }
+
+	    if (wl) {
+		/* Rewrite the last "..." to ",". */
+		wl -= 3;
+		p[wl++] = ',';
+	    }
+	    wl += snpf(p + wl, len - wl,
+		       (is_last_item? "%d": "%d..."), tfds[i]);
+	}
+
+	return wl;
+}
+
+static int
+fd_compare(const void *a, const void *b)
+{
+	int ia = *(int *)a, ib = *(int *)b;
+
+	return ia - ib;
+}
+
+static void
+snp_eventpoll(char *p, int len, int *tfds, int tfd_count)
+{
+	/* Reserve the area for prefix: "[eventpoll:" */
+	len -= 11;
+	/* Reserve the area for postfix */
+	len -= ((tfd_count == EPOLL_MAX_TFDS)
+		? 4 /* "...]" */
+		: 1 /* "]" */
+	       ) + 1 /* for the last \0 */
+	;
+
+	if (len > 1) {
+	    qsort (tfds, tfd_count, sizeof (tfds[0]), fd_compare);
+
+	    p[10] = ':'; /* "[eventpoll]" => "[eventpoll:" */
+	    int wl = snp_eventpoll_fds(p + 11, len, tfds, tfd_count);
+
+	    const char *postfix = (
+		/* If the buffer doesn't have enough space, snp_eventpoll_fds puts
+		 * "..." at the end of the buffer. In that case we don't
+		 have to "..." here. */
+		(wl > 3 && p [11 + wl - 1] == '.')
+		? "]"
+		: ((tfd_count == EPOLL_MAX_TFDS)
+		   /* File descriptors more than EPOLL_MAX_TFDS are associated to
+		    * the eventpoll fd. */
+		   ? "...]"
+		   : "]"));
+	    strcpy(p + 11 + wl, postfix);
+	}
+}
