@@ -1,5 +1,5 @@
 /*
- * dnode.c - NEXTSTEP and OpenStep node functions for lsof
+ * dnode.c - NEXTSTEP and OPENSTEP node functions for lsof
  */
 
 
@@ -32,7 +32,6 @@
 #ifndef lint
 static char copyright[] =
 "@(#) Copyright 1994 Purdue Research Foundation.\nAll rights reserved.\n";
-static char *rcsid = "$Id: dnode.c,v 1.13 2000/12/01 13:09:09 abe Exp $";
 #endif
 
 
@@ -197,7 +196,7 @@ load_svnc()
 		(void) fprintf(stderr,
 		    "%s: no space for %d local shadow vnode hash buckets\n",
 		    Pn, LF_SVNODE_HSZ);
-		Exit(1);
+		Error();
 	    }
 	}
 /*
@@ -224,7 +223,7 @@ load_svnc()
 		    (void) fprintf(stderr,
 			"%s: no space for local shadow vnode -- PID: %ld\n",
 			Pn, Lp->pid);
-		    Exit(1);
+		    Error();
 		}
 		lsv->vp = (KA_T)sv.lf_vnodep;
 		lsv->lp = (struct l_lockf *)NULL;
@@ -246,7 +245,7 @@ load_svnc()
 			(void) fprintf(stderr,
 			    "%s: no space for local lock struct -- PID: %ld\n",
 			    Pn, Lp->pid);
-			Exit(1);
+			Error();
 		    }
 		    lsf->type = lf.lf_type;
 		    lsf->start = lf.lf_start;
@@ -334,7 +333,7 @@ process_node(va)
 #endif	/* defined(HAS_AFS) */
 
 			      );
-		Exit(1);
+		Error();
 	    }
 	}
 	if (readvnode(va, v)) {
@@ -381,7 +380,7 @@ process_node(va)
 	/*
 	 * Caution: this AFS test should be the last one.
 	 */
-	
+
 	else if (avops) {
 	    if ((KA_T)v->v_op == avops)
 		Ntype = N_AFS;
@@ -468,10 +467,10 @@ unknown_v_op:
 	default:
 
 	/*
-	 * VCHR and VFIFO vnodes point to an snode.  The snode's s_realvp
+	 * VBLK, VCHR and VFIFO vnodes point to an snode.  The snode's s_realvp
 	 * usually points to a real vnode, which points to an inode.
 	 */
-	    if (v->v_type == VCHR || v->v_type == VFIFO) {
+	    if (v->v_type == VBLK || v->v_type == VCHR || v->v_type == VFIFO) {
 		if (!v->v_data || readsnode((KA_T)v->v_data, &s)) {
 		    (void) snpf(Namech, Namechl,
 			"vnode at %s: can't read snode(%s)",
@@ -500,7 +499,7 @@ unknown_v_op:
 		    ins = 1;
 		}
 		break;
-	    } else if (v->v_type != VBLK) {
+	    } else {
 		if (!v->v_data || readinode((KA_T)v->v_data, &i)) {
 		    (void) snpf(Namech, Namechl,
 			"vnode at %s: can't read inode (%s)",
@@ -550,20 +549,20 @@ unknown_v_op:
 #if	defined(HAS_AFS)
 	case N_AFS:
 	    if (an.ino_st) {
-		Lf->inode = an.inode;
+		Lf->inode = (INODETYPE)an.inode;
 		Lf->inp_ty = 1;
 	    }
 	    break;
 #endif	/* defined(HAS_AFS) */
 
 	case N_NFS:
-	    Lf->inode = (unsigned long)r.r_attr.va_nodeid;
+	    Lf->inode = (INODETYPE)r.r_attr.va_nodeid;
 	    Lf->inp_ty = 1;
 	    break;
 	case N_FIFO:
 	case N_REGLR:
 	    if (ins) {
-		Lf->inode = (unsigned long)i.i_number;
+		Lf->inode = (INODETYPE)i.i_number;
 		Lf->inp_ty = 1;
 	    }
 	}
@@ -679,11 +678,7 @@ unknown_v_op:
 	    ty = "FIFO";
 	    break;
 	default:
-	    if (type > 9999)
-		(void) snpf(Lf->type, sizeof(Lf->type), "*%03d", type % 1000);
-	    else
-		(void) snpf(Lf->type, sizeof(Lf->type), "%4d", type);
-	    (void) snpf(Namech, Namechl, "unknown type");
+	    (void) snpf(Lf->type, sizeof(Lf->type), "%04o", (type & 0xfff));
 	    ty = NULL;
 	}
 	if (ty)
