@@ -32,7 +32,6 @@
 #ifndef lint
 static char copyright[] =
 "@(#) Copyright 1994 Purdue Research Foundation.\nAll rights reserved.\n";
-static char *rcsid = "$Id: ddev.c,v 1.16 2001/03/19 12:19:54 abe Exp $";
 #endif
 
 
@@ -43,7 +42,8 @@ static char *rcsid = "$Id: ddev.c,v 1.16 2001/03/19 12:19:54 abe Exp $";
  * Local definitions
  */
 
-#define	LIKE_NODE_TTL	"like device special "
+#define	LIKE_BLK_SPEC	"like block special"
+#define	LIKE_CHR_SPEC	"like character special"
 
 
 /*
@@ -85,17 +85,17 @@ make_devtp(s, p)
 	    if (!Devtp) {
 		(void) fprintf(stderr, "%s: no space for character device\n",
 		    Pn);
-		Exit(1);
+		Error();
 	    }
 	}
 /*
  * Store the device number, inode number, and name in the Devtp[] entry.
  */
-	Devtp[Devx].inode = s->st_ino;
+	Devtp[Devx].inode = (INODETYPE)s->st_ino;
 	if (!(Devtp[Devx].name = mkstrcpy(p, (MALLOC_S *)NULL))) {
 	    (void) fprintf(stderr, "%s: no space for /dev/", Pn);
 	    safestrprt(p, stderr, 1);
-	    Exit(1);
+	    Error();
 	}
 	Devtp[Devx].rdev = s->st_rdev;
 	Devtp[Devx].v = 0;
@@ -104,14 +104,15 @@ make_devtp(s, p)
 
 
 /*
- * printchdevname() - print character device name
+ * printdevname() - print block or character device name
  */
 
 int
-printchdevname(dev, rdev, f)
+printdevname(dev, rdev, f, nty)
 	dev_t *dev;			/* device */
 	dev_t *rdev;			/* raw device */
 	int f;				/* 1 = print trailing '\n' */
+	int nty;			/* node type: N_BLK or N_CHR */
 {
 	struct clone *c;
 	struct l_dev *dp;
@@ -128,28 +129,44 @@ printchdevname_again:
 
 #endif	/* defined(HASDCACHE) */
 
-	if ((dp = lkupdev(dev, rdev, 1, 0))) {
+#if	defined(HASBLKDEV)
+	if (nty == N_BLK)
+	    dp = lkupbdev(dev, rdev, 1, 0);
+	else
+#endif	/* defined(HASBLKDEV) */
+
+	dp = lkupdev(dev, rdev, 1, 0);
+	if (dp) {
 	    safestrprt(dp->name, stdout, f);
 	    return(1);
 	}
 /*
  * Search device table for a match without inode number and dev.
  */
-	if ((dp = lkupdev(&DevDev, rdev, 0, 0))) {
+
+#if	defined(HASBLKDEV)
+	if (nty == N_BLK)
+	    dp = lkupbdev(&DevDev, rdev, 0, 0);
+	else
+#endif	/* defined(HASBLKDEV) */
+
+	dp = lkupdev(&DevDev, rdev, 0, 0);
+	if (dp) {
 
 	/*
-	 * A raw device match was found.  Record it as a name column addition.
+	 * A match was found.  Record it as a name column addition.
 	 */
-	    char *cp;
+	    char *cp, *ttl;
 	    int len;
 
-	    len = (int)(1 + strlen(LIKE_NODE_TTL) + strlen(dp->name) + 1);
+	    ttl = (nty == N_BLK) ? LIKE_BLK_SPEC : LIKE_CHR_SPEC;
+	    len = (int)(1 + strlen(ttl) + 1 + strlen(dp->name) + 1);
 	    if (!(cp = (char *)malloc((MALLOC_S)(len + 1)))) {
-		(void) fprintf(stderr, "%s: no nma space for: (%s%s)\n",
-		    Pn, LIKE_NODE_TTL, dp->name);
-		Exit(1);
+		(void) fprintf(stderr, "%s: no nma space for: (%s %s)\n",
+		    Pn, ttl, dp->name);
+		Error();
 	    }
-	    (void) snpf(cp, len + 1, "(%s%s)", LIKE_NODE_TTL, dp->name);
+	    (void) snpf(cp, len + 1, "(%s %s)", ttl, dp->name);
 	    (void) add_nma(cp, len);
 	    (void) free((MALLOC_P *)cp);
 	    return(0);
@@ -157,26 +174,26 @@ printchdevname_again:
 /*
  * Search for clone parent.
  */
-	if (Lf->is_stream && Clone && (*dev == DevDev)) {
+	if ((nty == N_CHR) && Lf->is_stream && Clone && (*dev == DevDev)) {
 	    for (c = Clone; c; c = c->next) {
-		if (major(*rdev) == minor(c->cd.rdev)) {
+		if (GET_MAJ_DEV(*rdev) == GET_MIN_DEV(c->cd.rdev)) {
 
 #if	defined(HASDCACHE)
 		    if (DCunsafe && !c->cd.v && !vfy_dev(&c->cd))
-		    goto printchdevname_again;
+			goto printchdevname_again;
 #endif	/* defined(HASDCACHE) */
 
 		    safestrprt(c->cd.name, stdout, f);
 		    return(1);
-		}	
+		}
 	    }
 	}
 /*
  * Search for pseudo device match on major device only.
  */
-	if (*dev == DevDev) {
+	if ((nty == N_CHR) && *dev == DevDev) {
 	    for (p = Pseudo; p; p = p->next) {
-		if (major(*rdev) == major(p->pd.rdev)) {
+		if (GET_MAJ_DEV(*rdev) == GET_MAJ_DEV(p->pd.rdev)) {
 
 # if	defined(HASDCACHE)
 		    if (DCunsafe && !p->pd.v && vfy_dev(&p->pd))
@@ -221,7 +238,7 @@ read_clone()
 	MALLOC_S pl;
 	struct pseudo *p;
 	struct stat sb;
-	
+
 	if (Clone || Pseudo)
 	    return;
 /*
@@ -230,7 +247,7 @@ read_clone()
 	if (!(path = mkstrcat(DVCH_DEVPATH, -1, "/", 1, "pseudo ", -1, &pl))) {
 	    (void) fprintf(stderr, "%s: no space for %s/pseudo\n",
 		DVCH_DEVPATH, Pn);
-	    Exit(1);
+	    Error();
 	}
 	path[pl - 1] = '\0';
 	if (!(dfp = OpenDir(path))) {
@@ -265,7 +282,7 @@ read_clone()
 		(void) fprintf(stderr, "%s: no space for: ", Pn);
 		safestrprt(path, stderr, 0);
 		safestrprt(dp->d_name, stderr, 1);
-		Exit(1);
+		Error();
 	    }
 
 #if	defined(USE_STAT)
@@ -309,7 +326,7 @@ read_clone()
 		    (void) fprintf(stderr,
 			"%s: no space for network clone device: ", Pn);
 		    safestrprt(fp, stderr, 1);
-		    Exit(1);
+		    Error();
 		}
 	    /*
 	     * Allocate space for the path name.
@@ -317,12 +334,12 @@ read_clone()
 		if (!(c->cd.name = mkstrcpy(fp, (MALLOC_S *)NULL))) {
 		    (void) fprintf(stderr, "%s: no space for clone name: ", Pn);
 		    safestrprt(fp, stderr, 1);
-		    Exit(1);
+		    Error();
 		}
 	    /*
 	     * Save the inode and device numbers.  Clear the verify flag.
 	     */
-		c->cd.inode = sb.st_ino;
+		c->cd.inode = (INODETYPE)sb.st_ino;
 		c->cd.rdev = sb.st_rdev;
 		c->cd.v = 0;
 	    /*
@@ -342,7 +359,7 @@ read_clone()
 	/*
 	 * Save pseudo device information.
 	 */
-	    if (minor(sb.st_rdev) == 0) {
+	    if (GET_MIN_DEV(sb.st_rdev) == 0) {
 
 	    /*
 	     * Allocate space for the pseduo device entry.
@@ -351,13 +368,13 @@ read_clone()
 		    (void) fprintf(stderr,
 			"%s: no space for pseudo device: ", Pn);
 		    safestrprt(fp, stderr, 1);
-		    Exit(1);
+		    Error();
 		}
 	    /*
 	     * Save the path name, and inode and device numbers.  Clear the
 	     * verify flag.  Link the entry to the pseudo chain.
 	     */
-		p->pd.inode = sb.st_ino;
+		p->pd.inode = (INODETYPE)sb.st_ino;
 		p->pd.name = fp;
 		fp = (char *)NULL;
 		p->pd.rdev = sb.st_rdev;
@@ -389,7 +406,6 @@ readdev(skip)
 #endif	/* defined(HASDCACHE) */
 
 	DIR *dfp;
-	int dn = 0;
 	struct DIRTYPE *dp;
 	char *fp = (char *)NULL;
 	MALLOC_S fpl;
@@ -420,13 +436,12 @@ readdev(skip)
 	    dcrd = 1;
 #endif	/* defined(HASDCACHE) */
 
-	Devx = 0;
 	if (!(ppath = mkstrcat(DVCH_DEVPATH, -1, "/", 1, "pseudo", -1,
 		      (MALLOC_S *)NULL)))
 	{
 	    (void) fprintf(stderr, "%s: no space for: %s/pseudo\n",
 		Pn, DVCH_DEVPATH);
-	    Exit(1);
+	    Error();
 	}
 	read_clone();
 	Dstk = (char **)NULL;
@@ -461,7 +476,7 @@ readdev(skip)
 	    {
 		(void) fprintf(stderr, "%s: no space for: ", Pn);
 		safestrprt(Dstk[Dstkx], stderr, 1);
-		Exit(1);
+		Error();
 	    }
 	    (void) free((FREE_P *)Dstk[Dstkx]);
 	    Dstk[Dstkx] = (char *)NULL;
@@ -484,7 +499,7 @@ readdev(skip)
 		    (void) fprintf(stderr, "%s: no space for: ", Pn);
 		    safestrprt(path, stderr, 0);
 		    safestrprt(dp->d_name, stderr, 1);
-		    Exit(1);
+		    Error();
 		}
 
 #if	defined(USE_STAT)
@@ -548,11 +563,11 @@ readdev(skip)
 			if (!BDevtp) {
 			    (void) fprintf(stderr,
 				"%s: no space for block device\n", Pn);
-			    Exit(1);
+			    Error();
 			}
 		    }
 		    BDevtp[j].rdev = sb.st_rdev;
-		    BDevtp[j].inode = sb.st_ino;
+		    BDevtp[j].inode = (INODETYPE)sb.st_ino;
 		    BDevtp[j].name = fp;
 		    fp = (char *)NULL;
 		    BDevtp[j].v = 0;
@@ -595,7 +610,7 @@ readdev(skip)
 	    {
 		(void) fprintf(stderr,
 		    "%s: no space for block device sort pointers\n", Pn);
-		Exit(1);
+		Error();
 	    }
 	    for (j = 0; j < BNdev; j++) {
 		BSdev[j] = &BDevtp[j];
@@ -621,7 +636,7 @@ readdev(skip)
 	    {
 		(void) fprintf(stderr,
 		    "%s: no space for character device sort pointers\n", Pn);
-		Exit(1);
+		Error();
 	    }
 	    for (i = 0; i < Ndev; i++) {
 		Sdev[i] = &Devtp[i];
@@ -631,7 +646,7 @@ readdev(skip)
 	    Ndev = rmdupdev(&Sdev, Ndev, 1);
 	} else {
 	    (void) fprintf(stderr, "%s: no character devices found\n", Pn);
-	    Exit(1);
+	    Error();
 	}
 
 #if	defined(HASDCACHE)
@@ -645,9 +660,6 @@ readdev(skip)
 }
 
 
-#if	defined(HASDCACHE)
-
-
 /*
  * clr_sect() - clear cached clone and pseudo sections
  */
@@ -680,6 +692,7 @@ clr_sect()
 }
 
 
+#if	defined(HASDCACHE)
 /*
  * rw_clone_sect() - read/write the device cache file clone section
  */
@@ -735,7 +748,7 @@ bad_clone_sect:
 		    (void) fprintf(stderr,
 			"%s: no space for cached clone: ", Pn);
 		    safestrprt(buf, stderr, 1);
-		    Exit(1);
+		    Error();
 		}
 	    /*
 	     * Enter the clone device number.
@@ -771,7 +784,7 @@ bad_clone_sect:
 	    /*
 	     * Enter the clone device inode number.
 	     */
-		for (c->cd.inode = (ino_t)0, ++cp; *cp != ' '; cp++) {
+		for (c->cd.inode = (INODETYPE)0, ++cp; *cp != ' '; cp++) {
 		    if (*cp < '0' || *cp > '9') {
 			if (!Fwarn) {
 			    (void) fprintf(stderr,
@@ -780,7 +793,8 @@ bad_clone_sect:
 			}
 			return(1);
 		    }
-		    c->cd.inode = (c->cd.inode * 10) + (ino_t)(*cp - '0');
+		    c->cd.inode = (INODETYPE)((c->cd.inode * 10)
+				+ (int)(*cp - '0'));
 		}
 	    /*
 	     * Enter the clone path name.
@@ -798,7 +812,7 @@ bad_clone_sect:
 		    (void) fprintf(stderr,
 			"%s: no space for cached clone path: ", Pn);
 		    safestrprt(buf, stderr, 1);
-		    Exit(1);
+		    Error();
 		}
 		c->cd.v = 0;
 		c->next = Clone;
@@ -836,7 +850,7 @@ bad_clone_sect:
  */
 	(void) fprintf(stderr, "%s: internal rw_clone_sect error: %d\n",
 	    Pn, m);
-	Exit(1);
+	Error();
 	return(1);		/* to make code analyzers happy */
 }
 
@@ -849,6 +863,8 @@ void
 rereaddev()
 {
 	(void) clr_devtab();
+	(void) clr_sect();
+	Devx = 0;
 
 # if	defined(DCACHE_CLR)
 	(void) DCACHE_CLR();
@@ -914,7 +930,7 @@ bad_pseudo_sect:
 		    (void) fprintf(stderr,
 			"%s: no space for cached pseudo: ", Pn);
 		    safestrprt(buf, stderr, 1);
-		    Exit(1);
+		    Error();
 		}
 	    /*
 	     * Enter the pseudo device number.
@@ -936,7 +952,7 @@ bad_pseudo_sect:
 	    /*
 	     * Enter the pseudo inode number.
 	     */
-		for (p->pd.inode = (ino_t)0; *cp != ' '; cp++) {
+		for (p->pd.inode = (INODETYPE)0; *cp != ' '; cp++) {
 		    if (*cp < '0' || *cp > '9') {
 			if (!Fwarn) {
 			    (void) fprintf(stderr,
@@ -945,7 +961,8 @@ bad_pseudo_sect:
 			}
 			return(1);
 		    }
-		    p->pd.inode = (p->pd.inode * 10) + (ino_t)(*cp - '0');
+		    p->pd.inode = (INODETYPE)((p->pd.inode * 10)
+				+ (int)(*cp - '0'));
 		}
 	    /*
 	     * Enter the pseudo path name.
@@ -967,7 +984,7 @@ bad_pseudo_sect:
 		    (void) fprintf(stderr,
 			"%s: no space for cached pseudo path: ", Pn);
 		    safestrprt(buf, stderr, 1);
-		    Exit(1);
+		    Error();
 		}
 		*(cp + len - 1) = '\0';
 		(void) snpf(p->pd.name, len, "%s", cp);
@@ -1033,7 +1050,7 @@ vfy_dev(dp)
 #endif	/* defined(USE_STAT) */
 
 	||  dp->rdev != sb.st_rdev
-	||  dp->inode != sb.st_ino) {
+	||  dp->inode != (INODETYPE)sb.st_ino) {
 	   (void) rereaddev();
 	   return(0);
 	}
@@ -1119,7 +1136,7 @@ rmdupdev(dp, n, ty)
 	{
 	    (void) fprintf(stderr, "%s: can't realloc %s device pointers\n",
 		Pn, ty ? "char" : "block");
-	    Exit(1);
+	    Error();
 	}
 	return(j);
 }
